package appmain;

import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import javax.microedition.io.*;
import java.io.*;

import menu.*;

/**
* A URLPage is used to display the content of a single URL.
* This is a dynamic Page, since it may take a while for the
* URL to load and we must return quickly from the constructor
* and doCommandAction.
*/
public class URLPage extends Page implements Runnable {
	/**
	* Initialize a new URLPage object using the supplied parameters.
	* @param label short label for this page (as might be used in
	* a menu)
	* @param title longer label for this page (as might be used in
	* a screen title)
	* @param icon small icon for this page (as might be used in
	* a toolbar or menu)
	* @param url the String describing the url to read
	*/
	public URLPage(String label,String title,Image icon,String url) {
		super(label,title,icon);
		if (url==null)
			throw new NullPointerException("URL must be supplied to "+this.getClass().getName());
			this.urlString = url;
	}
	/**
	* Initialize a new URLPage object using the supplied parameters
	* and default values where needed.
	* The title defaults to null (meaning that the label will be used
	* in its place).  
	* The icon defaults to null (meaning text only will be displayed).
	* @param label short label for this page (as might be used in
	* a menu)
	* @param url the String describing the url to read
	*/
	public URLPage(String label,String url) {
		this(label,null,null,url);
	}
	/**
	* Process the user command and return an integer value telling
	* the PageMgr what to do next.  This method handles the details
	* of a command for a page, while the commandAction method in the
	* PageMgr handles the general responsibility of switching displays.
	* @param c a Command object identifying the command.
	* @param d the Displayable on which this event has occurred
	* @return a pageAction code, selected from the public fields defined
	* by PageMgr.
	*/
	public int doCommandAction(Command c,Displayable d) {
		int pageAction;
		if (d != template) {
			throw new IllegalArgumentException(
			"Unknown Displayable received in "+
			this.getClass().getName());
		}
		synchronized (this) {
			if (c == PageMgr.BACK_COMMAND) {
				run = false;					// request content thread exit
				pageAction = PageMgr.BACK;
			} else if (c == PageMgr.CANCEL_COMMAND) {
				run = false;					// request content thread exit
				pageAction = PageMgr.BACK;
			} else {
				throw new IllegalArgumentException(
				"Unknown Command received in "+
				this.getClass().getName());
			}
		}
		return pageAction;
	}
	/**
	* Return the Displayable object for this URLPage, partially filled in
	* and ready to display.
	* @return the appropriate Displayable object for this Page
	*/
	public synchronized Displayable getDisplayable() {
		ticker.setString("Started");
		template.setTicker(ticker);
		template.setTitle(getLabel());
		((StringItem)template.get(0)).setText("-----");
		template.removeCommand(PageMgr.BACK_COMMAND);
		template.addCommand(PageMgr.CANCEL_COMMAND);
		return template;
	}
	/**
	* Do all the exciting connecting to the network and reading the contents
	* from the URL.  This method is started when the PageMgr sets this
	* page as the current displayable and then calls the start method.  The
	* method runs until completion or until the "run" variable is set to
	* false by the user command thread.  If run is set to false, then this
	* method must immediately clean up any outstanding resource holdings
	* that it has and return.  It must not touch anything that is shared
	* with other parts of the application because those objects (eg, the
	* screen display) are assuming that this thread is history and the 
	* shared objects may already be gone or in use by another thread.
	* Just clean up the local resource holdings and get out.
	* <p>Any action that affects objects that other threads may access or
	* change must be done within a synchronized block.  
	* <p>Any action that does not affect shared objects should not be done 
	* within a synchronized block so that the user command thread can dive
	* in and out of this object concurrently with this thread.
	* <p>Each little sychronized block is a unit that gets done
	* atomically.  Before starting the next one there's a chance 
	* that the user command thread will cancel this thread altogether by
	* setting run to false.
	* <p>Note that local variables in the run() method are allocated each
	* time a new thread starts execution in run().  Instance variables
	* are shared between all threads running in an object's run() method.
	*/
	public void run() {
		HttpConnection dC = null;		// the data Connection
		InputStream dS = null;			// the associated data Stream
		
		if (!nextStep("Opening connection",dC,dS)) return;
		try {							// open the connection
			dC = (HttpConnection)Connector.open(urlString);
		}
		catch (IOException e) {
			cleanUpForExit(e.getMessage(),dC,dS);
			return;
		}
		
		if (!nextStep("Opening input stream",dC,dS)) return;
		try {							// get input stream
			dS = dC.openInputStream();
		}
		catch (IOException e) {
			cleanUpForExit(e.getMessage(),dC,dS);
			return;
		}
		
		StringBuffer sbuf = new StringBuffer();
		int ch;
		
		if (!nextStep("Reading Data",dC,dS)) return;
		try {							// read the input stream
			while ((ch = dS.read()) != -1) {
				sbuf.append((char)ch);
			}
		}
		catch (IOException e) {
			cleanUpForExit(e.getMessage(),dC,dS);
			return;
		}
		
		synchronized (this) {			// tell the people what we saw
			if (run) {
				// template.setTicker(null);  // crash ... bug in emulator
				((StringItem)template.get(0)).setText(sbuf.toString());
			}
		}
		cleanUpForExit("Done",dC,dS);
	}
	/**
	* Helper method to condense the "check for termination" and
	* "display next step" functions.
	* @param m the message to display in the ticker if we are
	* going to go ahead with the next step.
	* @param c the Connection, either open or null
	* @param s the InputStream, either open or null
	* @return true if we should continue, false if this thread
	* should return immediately.
	*/
	private synchronized boolean nextStep(String m,HttpConnection c,InputStream s) {
		if (!run) {
			cleanUpForExit("",c,s);
			return false; 
		} else {
			ticker.setString(m);
			return true;
		}
	}
	/**
	* Helper method to prepare for thread termination.
	* Note that after this method executes the caller must 
	* immediately return from the run() method thereby terminating 
	* the thread.  This method makes a promise that this thread
	* is about to exit, and it is up to the caller to actually
	* follow through on the promise.
	* @param m the status message to display if we are still allowed
	* to do such things
	* @param c the Connection, either open or null
	* @param s the InputStream, either open or null
	*/
	private synchronized void cleanUpForExit(String m,HttpConnection c,InputStream s) {
		if (run) {
			ticker.setString(m);
			template.removeCommand(PageMgr.CANCEL_COMMAND);
			template.addCommand(PageMgr.BACK_COMMAND);
			run = false;
		}
		try { if (s != null) s.close(); }
		catch (IOException e) {}
		try { if (c != null) c.close(); }
		catch (IOException e) {}
		theThread = null;
		return;
	}
	
	//----------------------------------------------------------------------
	
	/** the template that is filled out to be our Displayable */
	private URLPageScreen template = new URLPageScreen();
	
	/** the Ticker that contains status info */
	private Ticker ticker = new Ticker("Idle");
	
	/** the URL to read. */
	private String urlString;
	
	//--------------------------------------------------------------------
	
	/**
	* A URLPageScreen is the Displayable (the view) for a
	* URLPage (the model).
	*/
	private class URLPageScreen extends Form {
		URLPageScreen() {
			super("URL Template");
			append(new StringItem(null,""));
		}
	}
}
